/*
 *  passing_variables.cpp
 *  CppTutorial
 *
 *  This example show various flavours of passing parameters to functions. 
 */

#include <iostream>
#include <string>

using std::string;
using std::cout;
using std::endl;

//pass by value
void someFunctionA(string baz) {
	//baz holds a copy of foo
	//This assignment has NO effect on foo
	baz = "Modified by someFunctionA.";
}

//pass by reference
void someFunctionB(string& baz) {
	//function has read/write access to foo
	baz = "Modified by someFunctionB.";
}

//pass by value-to-pointer
void someFunctionC(string* bazPtr) {
	//function has read/write access to
	//foo by dereferencing bazPtr
	(*bazPtr) = "Modified by someFunctionC.";

	//Changing the pointer has NO effect on foo
	bazPtr = NULL;
}

//pass by pointer-to-const
void someFunctionD(const string* bazPtr) {

	//function has read access to foo 
	//by dereferencing bazPtr
	if (bazPtr != NULL)
		cout << "someFunctionC reads: " << *bazPtr << endl;

	//This wouldn't compile:
	//(*bazPtr) = "Modified by someFunctionC";
	
	//Again, changing bazPtr, which is
	//a local variable of type const string* has no effect
	//on foo
	bazPtr = NULL;
}

//pass by const pointer-to-const
void someFunctionE(const string * const bazPtr) {

	//function has read access to foo 
	//by dereferencing bazPtr
	cout << "someFunctionC reads: " << *bazPtr << endl;

	//This wouldn't compile:
	//(*bazPtr) = "Modified by someFunctionC";
	
	//Since the pointer itself is also const,
	//this would not compile either:
	//bazPtr = NULL;
}

//pass by ref-to-const
//This is the preferred way to pass read-only parameters
void someFunctionF(const string& baz) {
	
	//function has read-only access to baz
	//NOTE: baz CANNOT be NULL
	cout << "someFunctionD reads: " << baz << endl;
	
	//This wouldn't compile:
	//baz = "some other value";
	
}

int main(int argc, char* argv[]) {

	string foo("Original Value");

	//Pass by value
	someFunctionA(foo);
	//Output: "Original Value"
	cout << foo << std::endl;
	
	//Pass by reference
	someFunctionB(foo);
	//Output: "Modified by someFunctionB"
	cout << foo << std::endl;
	
	//Pass by value-to-pointer
	someFunctionC(&foo);
	//Output: "Modified by someFunctionC"
	cout << foo << std::endl;	
	
	//Pass by pointer-to-const
	someFunctionD(&foo);	
	//Pass by const-point-to-const
	someFunctionE(&foo);
	//pass by ref-to-const
	someFunctionF(foo);
	
	return EXIT_SUCCESS;
}