/*
 *  Example_shared_ptr.cpp
 *  CppTutorialPart2
 *
 *  This example shows how to use shared_ptr. We use the soon-to-be-standardized tr1 implementation.
 * 
 */

#include <iostream>
#include "Person.h"
#include "Student.h"

#include <string>
#include <stdexcept>
#include <vector>

//NOTE: this requires GCC >= 4.2  or Visual Studio >= 2008 SP 1
#ifdef __MSVC__
	#include <memory>
#elif __GNUC__
	#include <tr1/memory>
#endif


//These are mainly for code brevity
using std::tr1::shared_ptr;
using std::string;
using std::cout;
using std::endl;
using std::vector;
using std::exception;


//Typedef our shared_ptr to write more readable code
typedef shared_ptr<string > StringRef;
typedef shared_ptr<Person > PersonRef;
typedef shared_ptr<Student > StudentRef;

//Throwing exception somewhere won't be a problem now
void someFunctionA() {
	
	//This shared_ptr is destroyed during unwinding of the stack 
	//and therefore the string on heap memory is properly cleaned up
	StringRef sRef(new string("My String"));
	
	throw std::exception();
	
}

//Prints out a vector full of shared_ptr holding instances of persons
void printoutPeople(const vector<PersonRef > people) {

	for (unsigned int i = 0; i<people.size(); i++) {
		cout << people.at(i)->saySomething() << endl;
	}
	
}

int main(int argc, char* argv[]) {
	
	//This shared_ptr automatically deletes your
	//allocated string once it's destroyed
	//Note that you should create a shared_ptr in its own statement!
	StringRef sRef(new string("String on heap."));
	
	//Use StringRef just like a pointer to a string (string*)
	cout << "The string '" << *sRef  << "' is " << sRef->size() << " characters long" << endl;

	//This is our proper example using exceptions
	try {
		someFunctionA();
	} catch (const std::exception&) {
		//Caught the exception, no memleak happened, thanks to shared_ptr!
	}
	
	//We can use shared_ptr also polymorphically
	PersonRef pRef(new Student("Mike", 123456));
	
	//And we can also just put them into STL containers...
	vector<PersonRef > people;
	people.push_back(pRef);
	//and reuse objects
	//This won't create memory leaks, as the original shared_ptr is still in the vector!
	pRef = PersonRef(new Person("John"));
	people.push_back(pRef);
	
	//This will print:
	//	Hi, my name is Mike and I am a student (nr.: 123456).
	//	Hi, my name is John
	printoutPeople(people);

	return EXIT_SUCCESS;
}