/*
 *  operator_overloading.cpp
 *  CppTutorialPart2
 *
 *  The example shows how to overload operators for
 *  a user-defined data type.
 *
 */


#include <iostream>
using std::cout;
using std::endl;

class my_vector {

  //public:
private:
  float x,y;

public:
  my_vector(void) : x(0.0f), y(0.0f) {}
  my_vector(float nx, float ny) : x(nx), y(ny) {}


  
  // overload "+" in class
  // usually we would make the method const, since it
  // does not change any class data items; however in this example
  // we would have ambiguity, since we would have "two versions" 
  // of the + operator if we made the member-func "const"...
  // in a "real" program, we would define only ONE operator+ function
  // (either "in-class" or "outside-class")
  my_vector operator+(const my_vector &second) /* const */ {
    cout << "operator+ in class" << endl;
    return my_vector(this->x + second.x, this->y + second.y);
  }
  
  // (=> friend fix) for private member access even outside class
  friend my_vector operator+(const my_vector &first, const my_vector &second);



  // prefix vs. postfix
  
  // prefix ++
  my_vector &operator++(void) {
    ++x; ++y; // we could also have written "x++; y++;"
    return *this;
  }

  // postfix ++
  // return vector with values before incrementing them!
  my_vector operator++(int dummy) {
    my_vector tmp(x,y);
    x++; y++; // we could also have written "++x; ++y;"
    return tmp;
  }


  // overload "<<" for printing coordinates of my_vector
  // through "friend"-declaration we also have access to private elements
  friend std::ostream &operator<<(std::ostream &out, const my_vector &v);
  
};

// outside class: only access to public members (but "friend fix"!)
std::ostream &operator<<(std::ostream &out, const my_vector &v) {
    out << "( " << (v.x) << " | " << (v.y) << " )";
    return out;
}


// overload "+" outside class; no access to private members (=> friend fix)
my_vector operator+(const my_vector &first, const my_vector &second) {
    cout << "operator+ outside class" << endl;
    return my_vector(first.x + second.x, first.y + second.y);
}



int main(int argc, char *argv[]) {

  my_vector v1(3.0f, -4.0f);
  my_vector v2(-6.0f, 5.0f);

  my_vector v3;
  my_vector v4;
  my_vector v5;

  cout << "v1 = " << v1 << endl;
  cout << "v2 = " << v2 << endl;
  cout << "v3 = " << v3 << endl;

  v3 = v1 + v2;
  cout << "v3 = " << v3 << endl;

  v4 = v1.operator+(v2); //needs in-class operator+()
  cout << "v4 = " << v4 << endl;  

  v5 = operator+(v1, v2);
  cout << "v5 = " << v5 << endl;  

  v1 = my_vector(1.0f, 4.0f);
  cout << endl;
  cout << "prefix ++ test" << endl;
  cout << "v1 = " << v1 << endl;
  cout << "++v1 = " << ++v1 << endl;
  cout << "v1 = " << v1 << endl;

  cout << endl;
  cout << "postfix ++ test" << endl;
  cout << "v1 = " << v1 << endl;
  cout << "v1++ = " << v1++ << endl;
  cout << "v1 = " << v1 << endl;

 
  
} //main()
